use std::{path::Path, process::exit, sync::LazyLock, time::Duration};

use chrono::Utc;
use protocol::BUFFER_LENGTH;
use s2n_quic::{
    Connection, Server,
    stream::{ReceiveStream, SendStream},
};
use tokio::{
    io::{AsyncReadExt, AsyncWriteExt},
    sync::{RwLock, broadcast},
    time::sleep,
};

use crate::ServerConfig;

const LOCAL_AUDIO_BUFFERING_TIMEOUT: Duration = Duration::from_millis(10);

#[derive(Debug)]
struct User {
    user_id: u64,
    local_audio_data_sender: broadcast::Sender<UseredAudioData>,
}

static ONLINE_USERS: LazyLock<RwLock<Vec<User>>> = LazyLock::new(|| RwLock::new(vec![]));

#[derive(Debug, Clone, Copy)]
struct UseredAudioData {
    user_id: u64,
    audio_data: f32,
}

impl User {
    async fn new(user_id: u64) -> broadcast::Receiver<UseredAudioData> {
        let (local_audio_data_sender, local_audio_data_receiver) =
            broadcast::channel(BUFFER_LENGTH * 16);
        let new_user = User {
            user_id,
            local_audio_data_sender,
        };

        ONLINE_USERS.write().await.push(new_user);
        local_audio_data_receiver
    }
}

pub async fn start(server_config: ServerConfig) {
    let mut server = Server::builder()
        .with_io(server_config.server_address.as_str())
        .unwrap()
        .with_tls((
            Path::new(&server_config.certificate_path),
            Path::new(&server_config.certificate_key_path),
        ))
        .unwrap()
        .start()
        .unwrap();
    let mut user_id = 0;
    while let Some(connection) = server.accept().await {
        tokio::spawn(handle_client(user_id, connection));
        user_id += 1;
    }
}

async fn handle_client(user_id: u64, mut connection: Connection) {
    println!(
        "Server Name = {}",
        connection.server_name().unwrap().unwrap().to_string()
    );

    println!(
        "Client Address = {}",
        connection.remote_addr().unwrap().to_string()
    );

    let stream = connection
        .accept_bidirectional_stream()
        .await
        .unwrap()
        .unwrap();
    let (receive_stream, send_stream) = stream.split();
    let local_audio_receiver = User::new(user_id).await;
    tokio::spawn(receive_audio_stream(user_id, receive_stream));
    tokio::spawn(send_audio_stream(
        user_id,
        send_stream,
        local_audio_receiver,
    ));
}

async fn receive_audio_stream(user_id: u64, mut receive_stream: ReceiveStream) {
    loop {
        match receive_stream.read_f32().await {
            Ok(received_audio_data) => {
                for online_user in ONLINE_USERS.read().await.iter() {
                    let usered_audio_data = UseredAudioData {
                        user_id,
                        audio_data: received_audio_data,
                    };
                    if let Err(err_val) =
                        online_user.local_audio_data_sender.send(usered_audio_data)
                    {
                        eprintln!("Error: Send Audio Data | Local | {}", err_val);
                        exit(0);
                    }
                }
            }
            Err(err_val) => {
                eprintln!("Error: Receive Audio Data | Remote | {}", err_val);
                break;
            }
        }
    }
}

async fn send_audio_stream(
    user_id: u64,
    mut send_stream: SendStream,
    mut local_audio_receiver: broadcast::Receiver<UseredAudioData>,
) {
    let mut usered_audio_buffer = vec![];
    let mut time = Utc::now();
    loop {
        if time + LOCAL_AUDIO_BUFFERING_TIMEOUT > Utc::now() {
            match local_audio_receiver.recv().await {
                Ok(usered_audio_data) => {
                    usered_audio_buffer.push(usered_audio_data);
                }
                Err(err_val) => {
                    eprintln!("Error: Receive Audio Data | Local | {}", err_val);
                    break;
                }
            }
        } else {
            let mixed_buffer = mixer(user_id, &mut usered_audio_buffer).await;

            for mixed_buffer_data in mixed_buffer {
                if let Err(err_val) = send_stream.write_f32(mixed_buffer_data).await {
                    eprintln!("Error: Send Audio Data | Remote | {}", err_val);
                }
            }

            usered_audio_buffer.clear();
            time = Utc::now();
        }
    }
}

async fn mixer(user_id: u64, usered_audio_buffer: &mut Vec<UseredAudioData>) -> Vec<f32> {
    usered_audio_buffer.sort_by(|first, second| first.user_id.cmp(&second.user_id));

    let mut parsed_audio_data = vec![];
    let mut inner = vec![];
    let mut current_user = user_id;

    while let Some(usered_audio) = usered_audio_buffer.pop() {
        if usered_audio.user_id == user_id {
            continue;
        } else {
            if current_user == usered_audio.user_id {
                inner.push(usered_audio.audio_data);
            } else {
                parsed_audio_data.push(inner.clone());
                inner.clear();
                current_user = usered_audio.user_id;
            }
        }
    }
    parsed_audio_data.push(inner.clone());

    let mut mixed_buffer = vec![];
    for parsed_audio_single_user_buffer in parsed_audio_data.iter() {
        for (i, parsed_audio_single_user_buffer_data) in
            parsed_audio_single_user_buffer.iter().enumerate()
        {
            match mixed_buffer.get(i) {
                Some(current_data) => {
                    mixed_buffer[i] = current_data + parsed_audio_single_user_buffer_data
                }
                None => mixed_buffer.push(*parsed_audio_single_user_buffer_data),
            }
        }
    }

    mixed_buffer
}
